#!/usr/bin/env python
# filename: ssh_tunnel.py


#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


# Borrowed heavily from Paramiko:
# https://github.com/paramiko/paramiko/blob/master/demos/forward.py


import getpass
import os
import socket
import select

try:
    import SocketServer
except ImportError:
    import socketserver as SocketServer

import argparse
import sys

import paramiko

SSH_PORT = 22
DEFAULT_PORT = 27017


class ForwardServer(socketserver.ThreadingTCPServer):
    daemon_threads = True
    allow_reuse_address = True


class Handler(socketserver.BaseRequestHandler):
    def handle(self):
        try:
            chan = self.ssh_transport.open_channel(
                "direct-tcpip",
                (self.chain_host, self.chain_port),
                self.request.getpeername(),
            )
        except Exception as e:
            print(
                "Incoming request to {}:{} failed: {}".format(
                    self.chain_host, self.chain_port, repr(e)
                )
            )
            return
        if chan is None:
            print(
                "Incoming request to {}:{} was rejected by the SSH server.".format(
                    self.chain_host, self.chain_port
                )
            )
            return

        print(
            "Connected!  Tunnel open {} -> {} -> {}".format(
                self.request.getpeername(),
                chan.getpeername(),
                (self.chain_host, self.chain_port),
            )
        )
        while True:
            r, w, x = select.select([self.request, chan], [], [])
            if self.request in r:
                data = self.request.recv(1024)
                if len(data) == 0:
                    break
                chan.send(data)
            if chan in r:
                data = chan.recv(1024)
                if len(data) == 0:
                    break
                self.request.send(data)
        peername = self.request.getpeername()
        chan.close()
        self.request.close()
        print("Tunnel closed from {}".format(peername,))


def forward_tunnel(local_port, remote_host, remote_port, transport):
    # this is convoluted, but necessary to configure things for the Handler
    # object.  (SocketServer doesn't give Handlers any way to access the outer
    # server.)
    class SubHander(Handler):
        chain_host = remote_host
        chain_port = remote_port
        ssh_transport = transport

    ForwardServer(("", local_port), SubHander).serve_forever()


def get_host_port(spec, default_port):
    "parse 'hostname:22' into a host and port, with the port optional"
    args = (spec.split(":", 1) + [default_port])[:2]
    args[1] = int(args[1])
    return args[0], args[1]


HELP = """\
Set up a forward tunnel across an SSH server, using paramiko. A local port
(given with -p) is forwarded across an SSH session to an address:port from
the SSH server. This is similar to the openssh -L option.
"""


def parse_arguments():
    parser = argparse.ArgumentParser(
        usage="usage: ssh_tunnel [options] <ssh-server>[:<server-port>]",
        version="ssh_tunnel",
        description=HELP,
    )
    parser.add_argument(
        "ssh_server", nargs=1, help="SSH server, as <ssh-server>[:<ssh-port>]"
    )
    parser.add_argument(
        "-p",
        "--local-port",
        action="store",
        type=int,
        dest="port",
        default=DEFAULT_PORT,
        help="local port to forward (default: {})".format(DEFAULT_PORT),
    )
    parser.add_argument(
        "-u",
        "--user",
        action="store",
        type=str,
        dest="user",
        default=getpass.getuser(),
        help="username for SSH authentication (default: {})".format(getpass.getuser()),
    )
    parser.add_argument(
        "-K",
        "--key",
        action="store",
        type=str,
        dest="keyfile",
        default=None,
        help="private key file to use for SSH authentication",
    )
    parser.add_argument(
        "--no-key",
        action="store_false",
        dest="look_for_keys",
        default=True,
        help="don't look for or use a private key file",
    )
    parser.add_argument(
        "-P",
        "--password",
        dest="readpass",
        default=False,
        action="store_true",
        help="Use a password for SSH.",
    )
    parser.add_argument(
        "-r",
        "--remote",
        action="store",
        required=True,
        type=str,
        dest="remote",
        default=None,
        metavar="host:port",
        help="remote host and port to forward to",
    )
    args = parser.parse_args()

    if args.remote is None:
        parser.error("Remote address required (-r).")

    server_host, server_port = get_host_port(args.ssh_server[0], SSH_PORT)
    remote_host, remote_port = get_host_port(args.remote, args.port)
    return args, (server_host, server_port), (remote_host, remote_port)


def main():
    args, server, remote = parse_arguments()
    if args.readpass:
        password = getpass.getpass("Enter SSH password: ")

    client = paramiko.SSHClient()
    client.load_system_host_keys()
    client.set_missing_host_key_policy(paramiko.WarningPolicy())

    print("Connecting to ssh host {}:{} ...".format(server[0], server[1]))
    try:
        client.connect(
            server[0],
            server[1],
            username=args.user,
            key_filename=args.keyfile,
            look_for_keys=args.look_for_keys,
            password=password,
        )
    except Exception as e:
        print("*** Failed to connect to {}:{}: {}".format(server[0], server[1], e))
        sys.exit(1)

    print("Now forwarding port {} to {}:{} ...".format(args.port, remote[0], remote[1]))
    try:
        forward_tunnel(args.port, remote[0], remote[1], client.get_transport())
    except KeyboardInterrupt:
        print("C-c: Port forwarding stopped.")
        sys.exit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# filename: sequence.py


#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import csv
import json
import operator
import os
import sys
import tempfile
import uuid
from collections import OrderedDict
from typing import Iterable, Optional, Union

import pandas as pd
from Bio import SeqIO
from Bio.SeqRecord import SeqRecord

from ..utils.codons import codon_lookup
from ..utils.pipeline import make_dir
from ..utils.utilities import nested_dict_lookup

STR_TYPES = [str]


class Sequence(object):
    """
    Container for biological (RNA, DNA, or protein) sequences.

    ``seq`` can be one of several things:

        1) a raw sequence, as a string

        2) an iterable, formatted as ``[seq_id, sequence]``

        3) a dict, containing at least the sequence ID and a
           sequence. Alternate ``id_key`` and ``seq_key``
           can be provided at instantiation.

        4) a Biopython ``SeqRecord`` object

        5) an abutils ``Sequence`` object

    If ``seq`` is provided as a string, the sequence ID can optionally be
    provided via the ``id`` keyword argument.  If ``seq`` is a string and ``id`` is not provided,
    a random sequence ID will be generated with ``uuid.uuid4()``.

    Quality scores can be supplied with ``qual`` or as part of a ``SeqRecord`` object.
    If providing both a SeqRecord object with quality scores and quality scores
    via ``qual``, the ``qual`` scores will override the SeqRecord quality scores.

    If ``seq`` is a dictionary, typically the result of a MongoDB query, the dictionary
    can be accessed directly from the ``Sequence`` instance (via the ``annotations`` property). To retrive the value
    for ``'junc_aa'`` in the instantiating dictionary, you would simply:

    .. code-block:: python

        s = Sequence(dict)
        junc = s['junc_aa']

    If ``seq`` is a dictionary, an optional ``id_key`` and ``seq_key`` can be provided,
    which tells the ``Sequence`` object which field to use to populate ``Sequence.id`` and
    ``Sequence.sequence``. Defaults for both ``id_key`` and ``seq_key`` are ``None``, which
    results in abutils trying to determine the appropriate key. For ``id_key``, the following
    keys are tried: ``['seq_id', 'sequence_id']``. For ``seq_key``, the following keys are tried:
    ``['vdj_nt', 'sequence_nt', 'sequence']``. If none of the attempts are successful, the
    ``Sequence.id`` or ``Sequence.sequence`` attributes will be ``None``.

    Alternately, the ``__getitem__()`` interface can be used to obtain a slice from the
    ``sequence`` attribute. An example of the distinction:

    .. code-block:: python

        d = {'name': 'MySequence', 'sequence': 'ATGC'}
        seq = Sequence(d, id_key='name', seq_key='sequence')

        seq['name']  # 'MySequence'
        seq[:2]  # 'AT'

    If the ``Sequence`` is instantiated with a dictionary, calls to ``__contains__()`` will
    return ``True`` if the supplied item is a key in the dictionary. In non-dict instantiations,
    ``__contains__()`` will look in the ``Sequence.sequence`` field directly (essentially a
    motif search). For example:

    .. code-block:: python

        dict_seq = Sequence({'seq_id': 'seq1', 'vdj_nt': 'ACGT'})
        'seq_id' in dict_seq  # TRUE
        'ACG' in dict_seq     # FALSE

        str_seq = Sequence('ACGT', id='seq1')
        'seq_id' in str_seq  # FALSE
        'ACG' in str_seq     # TRUE

    .. note::

        When comparing ``Sequence`` objects, they are comsidered equal only if their
        sequences and IDs are identical. This means that two ``Sequence`` objects
        with identical sequences but without user-supplied IDs won't be equal,
        because their IDs will have been randomly generated.
    """

    def __init__(
        self,
        seq: Union[str, Iterable, dict, SeqRecord],
        id: Optional[str] = None,
        qual: Optional[str] = None,
        annotations: Optional[dict] = None,
        id_key: Optional[str] = None,
        seq_key: Optional[str] = None,
    ):
        super(Sequence, self).__init__()
        self._input_sequence = None
        self._input_id = id
        self._input_qual = qual
        self._annotations = annotations
        self.id = None
        self.sequence = None
        self.qual = None
        self._fasta = None
        self._fastq = None
        self._reverse_complement = None
        self._strand = None
        self.id_key = id_key
        self.seq_key = seq_key
        self.description = None

        self._process_input(seq, id, qual)

    def __len__(self) -> int:
        """
        int: Returns the length of the ``sequence`` attribute
        """
        return len(self.sequence)

    def __iter__(self) -> Iterable[str]:
        """iter: Returns an iterator over the ``sequence`` attribute"""
        return iter(self.sequence)

    def __reversed__(self) -> str:
        """
        str: Returns the reverse of the ``sequence`` attribute
        """
        return "".join(reversed(self.sequence))

    def __contains__(self, item: str) -> bool:
        """
        Returns True if the supplied item is in the ``Sequence`` object.

        Parameters
        ----------

        item : str
            Item to be checked for membership in the ``Sequence`` object.

        Returns
        -------
        bool
            If the instance was initialzed with a dictonary (which means
            the ``annotations`` attribute is not empty), ``__contains__(key)``
            will return ``True`` if ``key`` is in ``annotations.keys()``. If ``annotations``
            is an empty dict, indicating instantiation without a dictionary,
            ``__contains__(motif)`` will return True if ``motif`` is in the
            ``sequence`` attribute.

        """
        if self.annotations:
            return item in self.annotations.keys()
        return item in self.sequence

    def __getitem__(self, key) -> Union[str, dict, None]:
        """
        Returns a slice of the ``Sequence`` object.

        Parameters
        ----------
        key : Union[int, str, slice]
            If ``Sequence`` was instantiated with a dictionary, ``key`` can be a string
            that corresponds to a key in the dictionary. If ``Sequence`` was instantiated
            with a string, ``key`` can be an integer or a slice.

        Returns
        -------
        Union[str, dict, None]
            If ``key`` is a string and the ``Sequence`` was instantiated with a dictionary,
            the value for the key will be returned. If ``key`` is a string and the ``Sequence``
            was instantiated with a string, ``None`` will be returned. If ``key`` is an integer
            or a slice, the corresponding slice of the ``sequence`` attribute will be returned.

        """
        if type(key) == slice:
            return self.sequence[key]
        elif key in self.annotations.keys():
            return self.annotations.get(key, None)
        elif isinstance(key, int):
            return self.sequence[key]
        return None

    def __setitem__(
        self, key: Union[str, int, float], val: Union[str, int, float]
    ) -> None:
        """
        Sets a value in the ``annotations`` property of a ``Sequence`` object.

        Parameters
        ----------
        key : Union[str, int, float]
            Key for the annotation to be set.

        val : Union[str, int, float]
            Value to be set for the annotation.

        """
        self.annotations[key] = val

    def __eq__(self, other: "Sequence") -> bool:
        """
        Returns True if the sequences and IDs of two ``Sequence`` objects are identical.

        Parameters
        ----------
        other : Sequence
            Another ``Sequence`` object.

        """
        if all([hasattr(other, "sequence"), hasattr(other, "id")]):
            return all([self.sequence == other.sequence, self.id == other.id])
        return False

    @classmethod
    def from_json(cls, data):
        if isinstance(data, str):
            data = json.loads(data)
        s = cls(None)
        s.__dict__ = data
        return s

    @property
    def fasta(self) -> str:
        """
        Returns the sequence, as a FASTA-formatted string.

        .. note::

                The FASTA string is built using ``Sequence.id`` and ``Sequence.sequence``.

        Returns
        -------
        str
            Returns the sequence, as a FASTA-formatted string.

        """
        if not self._fasta:
            self._fasta = ">{}\n{}".format(self.id, self.sequence)
        return self._fasta

    @property
    def fastq(self):
        """
        Returns the sequence, as a FASTQ-formatted string.

        .. note::

            The FASTQ string is built using ``Sequence.id``, ``Sequence.sequence``, and
            ``Sequence.qual``.

        Returns
        -------
        str
            Returns the sequence, as a FASTQ-formatted string

        """
        if self.qual is None:
            self._fastq = None
        else:
            if self._fastq is None:
                self._fastq = "@{}\n{}\n+\n{}".format(self.id, self.sequence, self.qual)
        return self._fastq

    @property
    def reverse_complement(self) -> Union[str, "Sequence"]:
        """
        Returns the reverse complement of ``Sequence.sequence``.

        Returns
        -------
            Returns a ``str`` if ``in_place`` is ``False``, otherwise returns an updated
            ``Sequence`` object in which the ``sequence`` property has been replaced
             with the reverse complement.

        """
        return reverse_complement(self, in_place=False)
        # if self._reverse_complement is None:
        #     self._reverse_complement = self._get_reverse_complement()
        # return self._reverse_complement

    @property
    def annotations(self):
        """
        Annotations is a dictionary that contains any additional information
        about the sequence. This can include sequence annotations, such as
        VDJ annotations, or any other information that might be useful to
        store with the sequence (like donor, group, etc).

        """
        if self._annotations is None:
            self._annotations = {}
        return self._annotations

    @annotations.setter
    def annotations(self, val):
        self._annotations = val

    @property
    def strand(self):
        if self._strand is None:
            self._strand = "plus"
        return self._strand

    @strand.setter
    def strand(self, strand):
        self._strand = strand

    def translate(self, sequence_key: Optional[str] = None, frame: int = 1) -> str:
        """
        Translate a nucleotide sequence.

        Parameters
        ----------
        sequence_key : str, default=None
            Name of the annotation field containg the sequence to be translated.
            If not provided, ``Sequence.sequence`` is used.

        frame : int, default=1
            Reading frame to translate. Default is ``1``.

        Returns
        -------
        str
            Translated sequence

        """
        return translate(self, sequence_key, frame)

    def as_fasta(
        self, name_field: Optional[str] = None, seq_field: Optional[str] = None
    ) -> str:
        """
        Returns the sequence, as a FASTA-formatted string.

        Parameters
        ----------
        name_field : str, default=None
            Name of the annotation field containing the sequence name. If not provided,
            ``Sequence.id`` is used.

        seq_field : str, default=None
            Name of the annotation field containing the sequence. If not provided,
            ``Sequence.sequence`` is used.

        Returns
        -------
        str
            Returns the sequence, as a FASTA-formatted string.

        """
        name = None
        sequence = None
        if name_field is not None:
            name = self.annotations[name_field]
        if seq_field is not None:
            sequence = self.annotations[seq_field]
        if name is None:
            name = self.id
        if sequence is None:
            sequence = self.sequence
        return ">{}\n{}".format(name, sequence)

    def region(self, start=0, end=None):
        """
        Returns a region of ``Sequence.sequence``, in FASTA format.

        If called without kwargs, the entire sequence will be returned.

        Parameters
        ----------
        start : int, default=0
            Start position of the region to be returned. Default is 0.

        end : int, default=None
            End position of the region to be returned. Negative values

        Returns
        -------
        str
            A region of ``Sequence.sequence``, in FASTA format

        """
        if end is None:
            end = len(self.sequence)
        return ">{}\n{}".format(self.id, self.sequence[start:end])

    def keys(self):
        """
        Returns the keys of the ``annotations`` attribute.
        """
        return self.annotations.keys()

    def values(self):
        """
        Returns the values of the ``annotations`` attribute.
        """
        return self.annotations.values()

    def get(
        self, key: Union[str, int, float], default: Union[str, int, float, None] = None
    ) -> Union[str, int, float, None]:
        """
        Returns the value of a key in the ``annotations`` attribute.

        Parameters
        ----------
        key : Union[str, int, float]
            Key for the annotation to be returned.

        default : Union[str, int, float, None], default=None
            Value to be returned if the key is not in the ``annotations`` attribute.

        Returns
        -------
        Union[str, int, float, None]
            Value of the key in the ``annotations`` attribute.

        """
        return self.annotations.get(key, default)

    def _process_input(self, seq, id, qual):
        if type(seq) in STR_TYPES:
            if id is None:
                id = uuid.uuid4()
            self.sequence = str(seq).upper()
            self.id = id
            self.qual = qual
            self._input_sequence = self.sequence
        elif type(seq) == Sequence:
            self.id = id if id is not None else seq.id
            self.sequence = seq.sequence
            self.description = seq.description
            self.qual = seq.qual
            self._input_sequence = self.sequence
            self._annotations = seq._annotations
        elif type(seq) in [list, tuple]:
            self.id = str(seq[0])
            self.sequence = str(seq[1]).upper()
            self.qual = qual
            self._input_sequence = self.sequence
        elif type(seq) == SeqRecord:
            if qual is None:
                if "phred_quality" in seq.letter_annotations:
                    qual = "".join(
                        chr(q + 33) for q in seq.letter_annotations["phred_quality"]
                    )
                elif "solexa_quality" in seq.letter_annotations:
                    qual = "".join(
                        chr(q + 64) for q in seq.letter_annotations["solexa_quality"]
                    )
            self.id = id if id is not None else str(seq.id)
            self.description = str(seq.description)
            self.sequence = str(seq.seq).upper()
            self.qual = qual
            self._input_sequence = self.sequence
        elif type(seq) in [dict, OrderedDict]:
            if self.id_key is None:
                id_options = ["seq_id", "sequence_id"]
                if any([k in seq for k in id_options]):
                    self.id_key = [k for k in id_options if k in seq][0]
            if self.seq_key is None:
                seq_options = ["vdj_nt", "sequence_nt", "sequence"]
                if any([k in seq for k in seq_options]):
                    self.seq_key = [k for k in seq_options if k in seq][0]
            self.id = id if id is not None else seq.get(self.id_key, None)
            self.sequence = seq.get(self.seq_key, None)
            self.qual = qual
            self._input_sequence = self.sequence
            for k in seq:
                if k not in self.annotations:
                    self.annotations[k] = seq[k]


def reverse_complement(
    sequence: Union[str, "Sequence"], in_place: bool = False
) -> Union[str, "Sequence"]:
    """
    Returns the reverse complement of a nucleotide sequence.

    Parameters
    ----------
    sequence : Union[str, Sequence]
        Nucleotide sequence to be reverse complemented.

    in_place : bool, default=False
        If ``True``, the input sequence will be reverse complemented in place.
        If ``False``, the reverse complemented sequence will be returned as a ``str``.


    Returns
    --------
    Union[str, Sequence]
        If ``in_place`` is ``False``, the reverse complemented sequence will be returned as a ``str``.
        If ``in_place`` is ``True``, the input sequence will be reverse complemented in place.

    """
    s = Sequence(sequence)
    rc_dict = {
        "A": "T",
        "C": "G",
        "G": "C",
        "T": "A",
        "Y": "R",
        "R": "Y",
        "S": "S",
        "W": "W",
        "K": "M",
        "M": "K",
        "B": "V",
        "D": "H",
        "H": "D",
        "V": "B",
        "N": "N",
    }
    rc = "".join([rc_dict.get(res, res) for res in s.sequence[::-1]])
    if in_place:
        s.sequence = rc
        return s
    return rc


def translate(
    sequence: Sequence, sequence_key: Optional[str] = None, frame: int = 1
) -> str:
    """
    Translates a nucleotide sequence.

    Parameters
    ----------
    sequence : Sequence
        ``Sequence`` object to be translated. Required.

    sequence_key : str, default=None
        Name of the annotation field containg the sequence to be translated.
        If not provided, ``sequence.sequence`` is used.

    frame : int, default=1
        Reading frame to translate. Default is ``1``.


    Returns
    -------
    translated : str

    """
    if sequence_key is not None:
        seq = Sequence(
            nested_dict_lookup(sequence.annotations, sequence_key.split("."))
        )
    else:
        seq = Sequence(sequence)
    if seq is None:
        return None
    start = (frame % 3) - 1
    end = len(seq) - (len(seq[start:]) % 3)
    seq = seq[start:end]
    translated = ""
    for i in range(0, len(seq), 3):
        codon = seq[i : i + 3]
        if all([c == "-" for c in codon]):
            aa = "-"
        else:
            aa = codon_lookup.get(codon, "X")
        translated += aa
    return translated


def read_json(
    json_file: str,
    match: Optional[dict] = None,
    fields: Iterable = None,
    id_key: str = "seq_id",
    sequence_key: str = "vdj_nt",
) -> Iterable[Sequence]:
    """
    Reads a JSON-formatted annotation file and returns ``Sequence`` objects.

    Parameters
    ----------
    json_file : str
        Path to the input JSON file. Required.

    match : dict, default=None
        A ``dict`` for filtering sequences from the input file.
        Sequences must match all conditions to be returned. For example,
        the following ``dict`` will filter out all sequences for which
        the ``'locus'`` field is not ``'IGH'``:

        .. code-block:: python

            {'locus': 'IGH'}

    fields : list, default=None
        A ``list`` of fields to be retained in the output ``Sequence``
        objects. Fields must be column names in the input file.

    id_key : str, default="seq_id"
        Name of the annotation field containing the sequence ID. Used to
        populate the ``Sequence.id`` property.

    sequence_key : str, default="vdj_nt"
        Name of the annotation field containg the sequence. Used to
        populate the ``Sequence.sequence`` property.


    Returns
    -------
    sequences : list of ``Sequences``

    """
    if match is None:
        match = {}
    if fields is not None:
        if id_key not in fields:
            fields.append(id_key)
        if sequence_key not in fields:
            fields.append(sequence_key)
    sequences = []
    with open(json_file, "r") as f:
        for line in f:
            if not line.strip():
                continue
            j = json.loads(line.strip())
            try:
                if all(
                    [nested_dict_lookup(j, k.split(".")) == v for k, v in match.items()]
                ):
                    if fields is not None:
                        _fields = [f for f in fields if f in j]
                        j = {f: j[f] for f in _fields}
                    sequences.append(Sequence(j, id_key=id_key, seq_key=sequence_key))
            except KeyError:
                continue
    return sequences


def read_csv(
    csv_file: str,
    delimiter: str = ",",
    match: Optional[dict] = None,
    fields: Iterable = None,
    id_key: str = "sequence_id",
    sequence_key: str = "sequence",
) -> Iterable[Sequence]:
    """
    Reads a tabular annotation file and returns ``Sequence`` objects.

    Parameters
    ----------
    csv_file : str
        Path to the input CSV file. Required.

    delimiter : str, default=","
        Column delimiter. Default is ``","``.

    match : dict, default=None
        A ``dict`` for filtering sequences from the input file.
        Sequences must match all conditions to be returned. For example,
        the following ``dict`` will filter out all sequences for which
        the ``'locus'`` field is not ``'IGH'``:

        .. code-block:: python

            {'locus': 'IGH'}

    fields : list, default=None
        A ``list`` of fields to be retained in the output ``Sequence``
        objects. Fields must be column names in the input file.

    id_key : str, default="sequence_id"
        Name of the annotation field containing the sequence ID. Used to
        populate the ``Sequence.id`` property.

    sequence_key : str, default="sequence"
        Name of the annotation field containg the sequence. Used to
        populate the ``Sequence.sequence`` property.


    Returns
    -------
    sequences : list of ``Sequences``

    """
    if match is None:
        match = {}
    if fields is not None:
        if id_key not in fields:
            fields.append(id_key)
        if sequence_key not in fields:
            fields.append(sequence_key)
    sequences = []
    df = pd.read_csv(csv_file, delimiter=delimiter)
    for _, r in df.iterrows():
        r = r.to_dict()
        try:
            if all([r[k] == v for k, v in match.items()]):
                if fields is not None:
                    _fields = [f for f in fields if f in r]
                    r = {f: r[f] for f in _fields}
                sequences.append(Sequence(r, id_key=id_key, seq_key=sequence_key))
        except KeyError:
            continue
    return sequences


def read_airr(
    tsv_file: str, match: Optional[dict] = None, fields: Optional[Iterable] = None
) -> Iterable[Sequence]:
    """
    Reads an AIRR-formatted annotation file and returns ``Sequence`` objects.

    Parameters
    ----------
    tsv_file : str
        Path to the input TSV file. Required.

    match : dict, default=None
        A ``dict`` for filtering sequences from the input file.
        Sequences must match all conditions to be returned. For example,
        the following ``dict`` will filter out all sequences for which
        the ``'locus'`` field is not ``'IGH'``:

        .. code-block:: python

            {'locus': 'IGH'}

    fields : list, default=None
        A ``list`` of fields to be retained in the output ``Sequence``
        objects. Fields must be column names in the input file.


    Returns
    -------
    sequences : list of ``Sequences``
    """
    return read_csv(tsv_file, delimiter="\t", match=match, fields=fields)


def read_fasta(fasta: str) -> Iterable[Sequence]:
    """
    Reads FASTA-formatted sequence data and returns ``Sequence`` objects.

    Parameters
    ----------
    fasta : str
        Either a FASTA-formatted string or the path to a FASTA file. Required.


    Returns
    -------
    sequences : list of ``Sequences``

    """
    if os.path.isfile(fasta):
        with open(fasta) as f:
            sequences = [Sequence(s) for s in SeqIO.parse(f, "fasta")]
    else:
        from io import StringIO

        f = StringIO(fasta)
        sequences = [Sequence(s) for s in SeqIO.parse(f, "fasta")]
    return sequences


def read_fastq(fastq: str) -> Iterable[Sequence]:
    """
    Reads FASTQ-formatted sequence data and returns ``Sequence`` objects.
    Gzipped files are supported.

    Parameters
    ----------
    fastq : str
        Either a FASTQ-formatted string or the path to a FASTQ file. Required.


    Returns
    -------
    sequences : list of ``Sequences``

    """
    if os.path.isfile(fastq):
        if fastq.endswith(".gz"):
            import gzip

            open_func = gzip.open
        else:
            open_func = open
        with open_func(fastq, "rt") as f:
            sequences = [Sequence(s) for s in SeqIO.parse(f, "fastq")]
    else:
        from io import StringIO

        f = StringIO(fastq)
        sequences = [Sequence(s) for s in SeqIO.parse(f, "fastq")]
    return sequences


def from_mongodb(db, collection, match=None, project=None, limit=None):
    if match is None:
        match = {}
    if project is None:
        project = {}
    if limit is not None:
        results = db[collection].find(match, project, limit=limit)
    else:
        results = db[collection].find(match, project)
    return [Sequence(r) for r in results]


def to_fasta(
    sequences: Union[str, Iterable[Sequence], Iterable[SeqRecord], Iterable[Iterable]],
    fasta_file: Optional[str] = None,
    as_string: bool = False,
    id_key: Optional[str] = None,
    sequence_key: Optional[str] = None,
    tempfile_dir: str = "/tmp",
) -> str:
    """
    Writes sequences to a FASTA-formatted file or returns a FASTA-formatted string.

    Parameters
    ----------
    sequences : Iterable[Sequence]
        An iterable of any of the following:
            1. list of abutils ``Sequence`` objects
            2. FASTA-formatted string
            3. path to a FASTA-formatted file
            4. list of BioPython ``SeqRecord`` objects
            5. list of lists/tuples, of the format ``[sequence_id, sequence]``
        Required.

    fasta_file : str, default=None
        Path to the output FASTA file. If not provided and `as_string` is ``False``,
        a file will be created using ``tempfile.NamedTemporaryFile()``.

    as_string : bool, default=False
        Return a FASTA-formatted string rather than writing to file.

    id_key : str, default=None
        Name of the annotation field containing the sequence ID. If not provided,
        ``sequence.id`` is used.

    sequence_key : str, default=None
        Name of the annotation field containg the sequence. If not provided,
        ``sequence.sequence`` is used.

    tempfile_dir : str, default="/tmp"
        If `fasta_file` is not provided, directory into which the tempfile
        should be created. If the directory does not exist, it will be
        created. Default is "/tmp".


    Returns
    --------
    fasta : str
        Path to a FASTA file or a FASTA-formatted string

    """
    if isinstance(sequences, str):
        # if sequences is already a FASTA-formatted file
        if os.path.isfile(sequences):
            if as_string:
                with open(sequences, "r") as f:
                    fasta_string = f.read()
                return fasta_string
            return sequences
        # if sequences is a FASTA-formatted string
        else:
            fasta_string = sequences
    # if sequences is a list of Sequences
    elif all([type(s) == Sequence for s in sequences]):
        ids = [s.get(id_key, s.id) if id_key is not None else s.id for s in sequences]
        seqs = [
            s.get(sequence_key, s.sequence) if sequence_key is not None else s.sequence
            for s in sequences
        ]
        fasta_string = "\n".join(f">{i}\n{s}" for i, s in zip(ids, seqs))
    # anything else..
    else:
        fasta_string = "\n".join(
            [Sequence(s, id_key=id_key, seq_key=sequence_key).fasta for s in sequences]
        )
    # output
    if as_string:
        return fasta_string
    if fasta_file is None:
        make_dir(tempfile_dir)
        ff = tempfile.NamedTemporaryFile(dir=tempfile_dir, delete=False)
        ff.close()
        fasta_file = ff.name
    else:
        make_dir(os.path.dirname(fasta_file))
    with open(fasta_file, "w") as f:
        f.write(fasta_string)
    return fasta_file


def to_fastq(
    sequences: Union[str, Iterable[Sequence], Iterable[SeqRecord], Iterable[Iterable]],
    fastq_file: Optional[str] = None,
    as_string: bool = False,
    id_key: Optional[str] = None,
    sequence_key: Optional[str] = None,
    tempfile_dir: str = "/tmp",
) -> str:
    """
    Writes sequences to a FASTQ-formatted file or returns a FASTQ-formatted string.

    Parameters
    ----------
    sequences : Iterable[Sequence]
        An iterable of any of the following:
            1. list of abutils ``Sequence`` objects
            2. FASTQ-formatted string
            3. path to a FASTQ-formatted file
            4. list of BioPython ``SeqRecord`` objects
            5. list of lists/tuples, of the format ``[sequence_id, sequence]``
        Required.

    fastq_file : str, default=None
        Path to the output FASTQ file. If not provided and `as_string` is ``False``,
        a file will be created using ``tempfile.NamedTemporaryFile()``.

    as_string : bool, default=False
        Return a FASTA-formatted string rather than writing to file.

    id_key : str, default=None
        Name of the annotation field containing the sequence ID. If not provided,
        ``sequence.id`` is used.

    sequence_key : str, default=None
        Name of the annotation field containg the sequence. If not provided,
        ``sequence.sequence`` is used.

    tempfile_dir : str, default="/tmp"
        If `fasta_file` is not provided, directory into which the tempfile
        should be created. If the directory does not exist, it will be
        created. Default is "/tmp".


    Returns
    --------
    fasta : str
        Path to a FASTA file or a FASTA-formatted string

    """
    if isinstance(sequences, str):
        # if sequences is already a FASTQ-formatted file
        if os.path.isfile(sequences):
            if as_string:
                with open(sequences, "r") as f:
                    fastq_string = f.read()
                return fastq_string
            return sequences
        # if sequences is a FASTQ-formatted string
        else:
            fastq_string = sequences
    # if sequences is a list of Sequences
    elif all([type(s) == Sequence for s in sequences]):
        if not all([s.qual is not None for s in sequences]):
            err = "Not all provided sequences have quality scores, which are required for FASTQ format."
            raise RuntimeError(err)
        ids = [s.get(id_key, s.id) if id_key is not None else s.id for s in sequences]
        seqs = [
            s.get(sequence_key, s.sequence) if sequence_key is not None else s.sequence
            for s in sequences
        ]
        quals = [s.qual for s in sequences]
        fastq_string = "\n".join(
            f"@{i}\n{s}\n+\n{q}" for i, s, q in zip(ids, seqs, quals)
        )
    # anything else..
    else:
        fastq_string = "\n".join(
            [Sequence(s, id_key=id_key, seq_key=sequence_key).fastq for s in sequences]
        )
    # output
    if as_string:
        return fastq_string
    if fastq_file is None:
        make_dir(tempfile_dir)
        ff = tempfile.NamedTemporaryFile(dir=tempfile_dir, delete=False)
        ff.close()
        fastq_file = ff.name
    else:
        make_dir(os.path.dirname(fastq_file))
    with open(fastq_file, "w") as f:
        f.write(fastq_string)
    return fastq_file



from collections import Counter
import itertools
import sys
from typing import Iterable, Union

import pandas as pd
import numpy as np

from Levenshtein import distance

import fastcluster as fc
from scipy.cluster.hierarchy import fcluster

from mnemonic import Mnemonic

from ..core.sequence import Sequence
from ..tools import cluster
from ..utils.utilities import nested_dict_lookup


def clonify(
    sequences: Iterable[Sequence],
    distance_cutoff: float = 0.32,
    shared_mutation_bonus: float = 0.65,
    length_penalty_multiplier: Union[int, float] = 2,
    group_by_v: bool = True,
    group_by_j: bool = True,
    preclustering: bool = False,
    preclustering_threshold: float = 0.65,
    vgene_key: str = "v_gene",
    jgene_key: str = "j_gene",
    cdr3_key: str = "cdr3_aa",
    muts_key: str = "v_mutations",
    preclustering_key: str = "cdr3_nt",
    muts_delimiter: str = "|",
    lineage_field: str = "lineage",
    lineage_size_field: str = "lineage_size",
    return_assignment_dict: bool = False,
) -> Union[dict, Iterable[Sequence]]:
    """
    Assigns BCR sequences to clonal lineages using the clonify_ [Briney16]_ algorithm.

    .. seealso::
       | Bryan Briney, Khoa Le, Jiang Zhu, and Dennis R Burton
       | Clonify: unseeded antibody lineage assignment from next-generation sequencing data.
       | *Scientific Reports* 2016. https://doi.org/10.1038/srep23901

    Parameters
    ----------
    sequences : Iterable[Sequence]
        Iterable of ``abutils.core.Sequence`` objects containing annotated sequence data.

    distance_cutoff : float, default=0.32
        Distance threshold for lineage clustering.

    shared_mutation_bonus : float, default=0.65
        Bonus applied for each shared V-gene mutation.

    length_penalty_multiplier : int, default=2
        Multiplier for the CDR3 length penalty. Default is ``2``, resulting in CDR3s that
        differ by ``n`` amino acids being penalized ``n * 2``.

    group_by_v : bool, default=True
        If ``True``, sequences are grouped by V-gene use prior to lineage assignment. This
        option is additive with ``group_by_j``. For example, if ``group_by_v == True`` and
        ``group_by_j == True``, sequences will be grouped by both V-gene and J-gene.

    group_by_j : bool, default=True
        If ``True``, sequences are grouped by J-gene use prior to lineage assignment. This
        option is additive with ``group_by_v``. For example, if ``group_by_v == True`` and
        ``group_by_j == True``, sequences will be grouped by both V-gene and J-gene.

    preclustering : bool, default=False
        If ``True``, V/J groups are pre-clustered on the `preclustering_field` sequence,
        which can potentially speed up lineage assignment and reduce memory usage.
        If ``False``, each V/J group is processed in its entirety without pre-clustering.

    preclustering_threshold : float, default=0.65
        Identity threshold for pre-clustering the V/J groups prior to lineage assignment.

    vgene_key : str, default='v_gene'
        Annotation field containing the V-gene name.

    jgene_key : str, default='j_gene'
        Annotation field containing the J-gene name.

    cdr3_key : str, default='cdr3_aa'
        Annotation field containing the CDR3 amino acid sequence.

    muts_key : str, default='v_mutations'
        Annotation field containing the V-gene mutations.

    preclustering_key : str, default='cdr3_nt'
        Annotation field on which to pre-cluster sequences.

    muts_delimiter : str, default='|'
        Delimiter used to separate mutations in the `muts_key` annotation field.

    lineage_field : str, default='lineage'
        Name of the lineage assignment field.

    lineage_size_field : str, default='lineage_size'
        Name of the lineage size field.

    return_assignment_dict : bool, default=False
        If ``True``, a dictionary linking sequence IDs to lineage names will be returned.
        If ``False``, the input ``anndata.AnnData`` object will be returned, with lineage
        annotations included.

    Returns
    -------
    output : ``Iterable[Sequence]`` or ``dict``
        By default (``return_assignment_dict == False``), a list of input sequences are
        returned with two additional annotations: the lineage name (added to `lineage_field`)
        and lineage size (added to `lineage_size_field`). If ``return_assignment_dict == True``,
        a ``dict`` mapping sequence IDs to lineage names is returned.


    .. _clonify: https://github.com/briney/clonify
    """
    # select the appropriate data fields
    # if annotation_format.lower() == "airr":
    #     vgene_key = "v_gene"
    #     jgene_key = "j_gene"
    #     cdr3_key = "cdr3_aa"
    #     muts_key = "v_mutations"
    # elif annotation_format.lower() == "json":
    #     vgene_key = "v_gene.gene"
    #     jgene_key = "j_gene.gene"
    #     cdr3_key = "cdr3_aa"
    #     muts_key = "var_muts_nt.muts"
    # else:
    #     error = "ERROR: "
    #     error += 'annotation_format must be either "airr" or "json", '
    #     err += f"but you provided {annotation_format}"
    #     print("\n")
    #     print(error)
    #     print("\n")
    #     sys.exit()

    # setup a list of required fields
    required_fields = ["v_gene", "j_gene", "cdr3", "mutations"]
    if preclustering:
        required_fields.append("preclustering")
    # group sequences by V/J genes
    vj_group_dict = {}
    heavies = [s for s in sequences if s["locus"] == "IGH"]
    for h in heavies:
        # build new Sequence objects using just the data we need
        s = Sequence(h.sequence, id=h.id)
        s["v_gene"] = nested_dict_lookup(h, vgene_key.split("."))
        s["j_gene"] = nested_dict_lookup(h, jgene_key.split("."))
        s["cdr3"] = nested_dict_lookup(h, cdr3_key.split("."))
        muts = h[muts_key]
        if any([pd.isnull(muts), muts is None]):
            s["mutations"] = []
        elif isinstance(muts, str):
            muts = muts.split(muts_delimiter)
        else:
            muts = list(muts)
            s["mutations"] = [m for m in muts if m.strip()]
        if preclustering:
            s["preclustering"] = nested_dict_lookup(h, preclustering_key.split("."))
        # skip sequences that don't have all required fields
        if any([s[v] is None for v in required_fields]):
            continue
        # group sequences by VJ gene use
        vj = ""
        if group_by_v:
            vj += s["v_gene"]
        if group_by_j:
            vj += s["j_gene"]
        if vj not in vj_group_dict:
            vj_group_dict[vj] = []
        vj_group_dict[vj].append(s)
    # assign lineages
    assignment_dict = {}
    mnemo = Mnemonic("english")
    for vj_group in vj_group_dict.values():
        # preclustering
        if preclustering:
            seq_dict = {s.id: s for s in vj_group}
            cluster_seqs = [Sequence(s[preclustering_key], id=s.id) for s in vj_group]
            clusters = cluster(cluster_seqs, threshold=preclustering_threshold)
            groups = [[seq_dict[i] for i in c.seq_ids] for c in clusters]
        else:
            groups = [
                vj_group,
            ]
        for group in groups:
            # singletons
            if len(group) == 1:
                seq = group[0]
                assignment_dict[seq.id] = "_".join(
                    mnemo.generate(strength=128).split()[:6]
                )
            # if not a singleton, cluster
            else:
                # calculate distance matrix
                dist_matrix = []
                for s1, s2 in itertools.combinations(group, 2):
                    d = pairwise_distance(
                        s1, s2, shared_mutation_bonus, length_penalty_multiplier
                    )
                    dist_matrix.append(d)
                # cluster
                linkage_matrix = fc.linkage(
                    dist_matrix, method="average", preserve_input=False
                )
                cluster_list = fcluster(
                    linkage_matrix, distance_cutoff, criterion="distance"
                )
                # rename clusters
                cluster_ids = list(set(cluster_list))
                cluster_names = {
                    c: "_".join(mnemo.generate(strength=128).split()[:6])
                    for c in cluster_ids
                }
                renamed_clusters = [cluster_names[c] for c in cluster_list]
                # assign sequences
                for seq, name in zip(vj_group, renamed_clusters):
                    assignment_dict[seq.id] = name
    # return just the assignment dict if requested
    if return_assignment_dict:
        return assignment_dict
    # otherwise, add the lineage name and size to the sequences
    lineage_size_dict = Counter(assignment_dict.values())
    for s in sequences:
        if l := assignment_dict.get(s.id, None) is not None:
            s[lineage_field] = l
            s[lineage_size_field] = lineage_size_dict.get(l, 1)
    return sequences


def pairwise_distance(
    s1: Sequence,
    s2: Sequence,
    shared_mutation_bonus: float = 0.65,
    length_penalty_multiplier: Union[int, float] = 2,
    cdr3_field: str = "cdr3",
    mutations_field: str = "mutations",
) -> float:
    """
    Computes length and mutation adjusted Levenshtein distance for a pair of sequences.

    Parameters
    ----------
    s1 : Sequence
        The first input sequence

    s2 : Sequence
        The second input sequence

    shared_mutation_bonus : float, optional
        The bonus for each shared mutation, by default 0.65

    length_penalty_multiplier : Union[int, float], optional
        Used to compute the penalty for differences in CDR3 length. The length
        difference is multiplied by `length_penalty_multiplier`, by default 2

    cdr3_field : str, optional
        Name of the field in `s1` and `s2` containing the CDR3 sequence,
        by default ``"cdr3"``

    mutations_field : str, optional
        Name of the field in `s1` and `s2` containing mutation information,
        by default ``"mutations"``

    Returns
    -------
    distance: float
        The adjusted Levenshtein distance between the two sequences
    """
    if len(s1[cdr3_field]) == len(s2[cdr3_field]):
        dist = sum([i != j for i, j in zip(s1[cdr3_field], s2[cdr3_field])])
    else:
        dist = distance(s1[cdr3_field], s2[cdr3_field])
    length_penalty = (
        abs(len(s1[cdr3_field]) - len(s2[cdr3_field])) * length_penalty_multiplier
    )
    length = min(len(s1[cdr3_field]), len(s2[cdr3_field]))
    shared_mutations = list(set(s1[mutations_field]) & set(s2[mutations_field]))
    mutation_bonus = len(shared_mutations) * shared_mutation_bonus
    score = (dist + length_penalty - mutation_bonus) / length
    return max(score, 0.001)  # distance values can't be negative
